module xn2ns;

private import std.string;
private import std.conv;

private import nscr.charset;
private import nscr.util;

private import option;
private import xnovel;

int convert(string xnSource, ref Option opt){
    if(xnSource.length >= 3 && xnSource[0] == 0xef){
        // Remove BOM
        xnSource = xnSource[3..$];
    }
    auto xn = new XNovel(xnSource);
    string nscript;
    string appTitle = xn._head._title;
    auto visitor = new XN2NSVisitor(opt.nonop);
    try{
        xn._body._contents.visit(visitor);
    }
    catch(Exception e){
        std.stdio.writeln(toNScrCharset(e.toString()));
        return 10;
    }
    debug std.stdio.writeln("visited!");
    string crlf = newline;
    // define
    nscript ~= std.array.join(["*define", "linepage", ""], crlf);
    debug std.stdio.writeln("defined!");
    // aliases
    nscript ~= std.array.join(visitor.defines, crlf);
    foreach(name; numaliases.keys.sort){
        nscript ~= "numalias " ~ name ~ "," ~ numaliases[name] ~ crlf;
    }
    foreach(name; straliases.keys.sort){
        nscript ~= "stralias " ~ name ~ "," ~ toNScrCharset(straliases[name]) ~ crlf;
    }
    // execute
    nscript ~= std.array.join(["", "game", "", "*start", ""], crlf);
    debug std.stdio.writeln("executed!");
    // contents
    nscript ~= visitor.executes;
    
    debug std.stdio.writeln("contens!");
    // output
    if(opt.print){
        std.stdio.writeln(nscript);
    }
    else{
        std.file.write(opt.output, nscript);
    }
    return 0;
}

private string XNVAR_PATTERN = "(tmp|save|sys):([_a-zA-Z][._a-zA-Z0-9]*)";
private typedef string embeded;

private class XN2NSVisitor : ContentsVisitor{
    private string[] _lines;
    private string[string] _channels;
    private bool _nonop;
    private string _maxlength = "nscr_input_maxlength";
    private string _zenkaku = "nscr_input_zenkaku";
    private bool _prevtext;
    
    this(bool nonop){
        _nonop = nonop;
    }
    
    string[] defines(){
        string[] result;
        foreach(channel, name; _channels){
            result ~= "numalias " ~ name ~ "," ~ channel;
        }
        return result;
    }
    
    string executes(){
        string crlf = newline;
        string result;
        foreach(line; _lines){
            result ~= line ~ crlf;
        }
        return result;
    }
    
    void visit(XNStopbgm command){
        _lines ~= "bgmstop";
        _prevtext = false;
    }

    void visit(Img command){
        if(command.id == "bg"){
            _lines ~= format("bg %s,1", enquote(command.src));
        }
        else{
            // Should not use this.
            _lines ~= format("ld %s,%s,1", idToPos(command.id), enquote(command.src));
        }
        _prevtext = false;
    }

    void visit(Fg command){
        string time = "1";
        if(command.time.length){
            time = format("10,%d", secondsToMillis(command.time));
        }
        if(command.id == "bg"){
            _lines ~= format("bg %s,%s", enquote(command.src), time);
        }
        else{
            string pos;
            if(command.hasPos()){
                pos = yoko_posToPos(command.yoko_pos);
            }
            else if(command.id.length){
                pos = idToPos(command.id);
            }
            else{
                throw new Exception("位置を特定できません");
            }
            if(command.src.length){
                _lines ~= format("ld %s,%s,%s", pos, enquote(command.src), time);
            }
            else{
                _lines ~= format("cl %s,%s", pos, time);
            }
        }
        _prevtext = false;
    }

    void visit(Audio command){
        bool bgm;
        if(command.id.length){
            bgm = command.id == "bgm";
        }
        else{
            if(command.src.length <= 4) throw new Exception("ファイル名が不正です");
            bgm = std.algorithm.endsWith(command.src, ".mp3") == 1;
        }
        
        if(bgm){
            if(command.loop) _lines ~= format("bgm %s", enquote(command.src));
            else _lines ~= format("bgmonce %s", enquote(command.src));
        }
        else{
            if(command.id.length == 0){
                if(command.loop) _lines ~= format("waveloop %s", enquote(command.src));
                else _lines ~= format("wave %s", enquote(command.src));
            }
            else{
                string channel = command.id;
                if(!isdigits(channel)){
                    if(!(channel in _channels)){
                        _channels[channel] = to!string(_channels.length);
                    }
                    channel = _channels[channel];
                }
                // Auto Playは整合性を取れないので無視。
                if(command.loop) _lines ~= format("dwaveloop %s,%s", channel, enquote(command.src));
                else _lines ~= format("dwave %s,%s", channel, enquote(command.src));
            }
        }
        _prevtext = false;
    }
    
    void visit(Br command){
        if(command.id.length){
            _lines ~= "*" ~ command.id;
        }
        // linepageとするので単なるbrは不要
        _prevtext = false;
    }
    
    void visit(Text command){
        debug std.stdio.writeln("visit Text");
        string text = toNScrCharset(replaceAllVars!(embeded)(command.text));
        if(_prevtext){
            _lines[$-1] ~= text;
        }
        else{
            _lines ~= text;
        }
        _prevtext = true;
    }
    
    void visit(Nav command){
        debug std.stdio.writeln("visit Nav");
        _lines ~= toNScrCharset(command._title.text);
        string select;
        foreach(link; command._links){
            select ~= toNScrCharset(enquote(link.text)) ~ ",*" ~ link.href[1..$] ~ ",";
        }
        _lines ~= "select " ~ select[0..$-1];
        _prevtext = false;
    }
    
    void visit(A command){
        if(command.cond.length){
            string cond = std.array.replace(std.array.replace(command.cond, "＞", ">"), "＜", "<");
            if(maybeStringExpr(cond)){
                cond = replaceAllVars!(string)(cond);
            }
            else{
                cond = replaceAllVars!(int)(cond);
            }
            _lines ~= format("if %s goto *%s", cond, command.href[1..$]);
        }
        else{
            _lines ~= "goto *" ~ command.href[1..$];
        }
        _prevtext = false;
    }
    
    void visit(Ruby command){
        debug std.stdio.writeln("visit Ruby");
        string text = toNScrCharset("(" ~ command._rb ~ "/" ~ command._rt ~ ")");
        if(_prevtext){
            _lines[$-1] ~= text;
        }
        else{
            _lines ~= text;
        }
        _prevtext = true;
    }
    
    void visit(XNClose command){
        _lines ~= "end";
        _prevtext = false;
    }
    
    void visit(XNWait command){
        if(command.time.length){
            _lines ~= "wait " ~ to!string(secondsToMillis(command.time));
        }
        else{
            _lines ~= "lrclick";
        }
        _prevtext = false;
    }
    
    void visit(Video command){
        _lines ~= format("movie %s", enquote(command.src));
        _prevtext = false;
    }
    
    void visit(XNLet command){
        string var;
        string text = command.text;
        if(!command.calc){
            var = replaceAllVars!(string)(command.name);
            text = replaceAllVars!(string)(text);
            if(text == command.text){
                text = enquote(text);
            }
        }
        else if(maybeStringExpr(text)){
            var = replaceAllVars!(string)(command.name);
            text = tr(replaceAllVars!(string)(text), "'", "\"");
        }
        else{
            var = replaceAllVars!(int)(command.name);
            // Consider "tmp:aa.xx + 1".
            text = replaceAllVars!(int)(text);
        }
        debug std.stdio.writeln("visit XNLet");
        _lines ~= toNScrCharset(format("mov %s,%s", var, text));
        _prevtext = false;
    }
    
    void visit(XNNop command){
        if(_nonop){
            return;
        }
        debug std.stdio.writeln("visit XNNop");
        _lines ~= toNScrCharset(";" ~ command.text);
        _prevtext = false;
    }
    
    void visit(XNStopAllse command){
        _lines ~= "stop";
        _prevtext = false;
    }
    
    void visit(Input command){
        debug std.stdio.writeln("visit Input");
        numaliases[_maxlength] = "64";
        numaliases[_zenkaku] = "0";
        string var = replaceAllVars!(string)(command.name);
        _lines ~= toNScrCharset(format("input %s,%s,%s,%s,%s", var, enquote(command.title), enquote(command.placeholder), _maxlength, _zenkaku));
        _prevtext = false;
    }
    
    void visit(Blank command){
        _lines ~= "";
    }
}

string[] locals;
string[] globals;

string replaceAllVars(T)(string str){
    static if(is(T == int)){
        return std.regex.replace!(xnvar2numvar)(str, std.regex.regex(XNVAR_PATTERN, "g"));
    }
    else static if(is(T == string)){
        return std.regex.replace!(xnvar2strvar)(str, std.regex.regex(XNVAR_PATTERN, "g"));;
    }
    else static if(is(T == embeded)){
        return std.regex.replace!(xnvar2strvar)(str, std.regex.regex("#\\{" ~ XNVAR_PATTERN ~ "\\}", "g"));;
    }
}

unittest{
    string var;
    var = std.regex.replace!(xnvar2numvar)("tmp:i0", std.regex.regex(XNVAR_PATTERN));
    assert(var == "%tmp_i0", var);
    var = std.regex.replace!(xnvar2strvar)("tmp:s0", std.regex.regex(XNVAR_PATTERN));
    assert(var == "$tmp_s0", var);
}

string xnvar2strvar(std.regex.RegexMatch!(string) matcher){
    return "$" ~ xnvar2nsvar(matcher);
}

string xnvar2numvar(std.regex.RegexMatch!(string) matcher){
    return "%" ~ xnvar2nsvar(matcher);
}

string xnvar2nsvar(std.regex.RegexMatch!(string) matcher){
    auto captures = matcher.captures();
    auto prefix = captures[1];
    auto name = tr(captures[2], ".", "_");
    auto fullname = prefix ~ "_" ~ name;
    if(fullname in numaliases) return fullname;
    if(prefix == "sys"){
        numaliases[fullname] = to!string(globals.length + 200);
        globals ~= fullname;
    }
    else{
        numaliases[fullname] = to!string(locals.length);
        locals ~= fullname;
    }
    return fullname;
}

string idToPos(string id){
    // TODO Consider humanorder.
    switch(id){
    case "0":
        return "l";
    case "1":
        return "c";
    case "2":
        return "r";
    default:
        throw new Exception("Unknown ID");
    }
}

string yoko_posToPos(int yoko_pos){
    // TODO Consider humanorder.
    switch(yoko_pos){
    case 1:
    case 4:
    case 7:
        return "l";
    case 2:
    case 5:
    case 8:
        return "c";
    case 3:
    case 6:
    case 9:
        return "r";
    default:
        throw new Exception("Unknown yoko_pos");
    }
}

unittest{
    assert(secondsToMillis("0.5") == 500);
    assert(secondsToMillis("0.333333") == 333);
    assert(secondsToMillis("1") == 1000);
}

int secondsToMillis(string seconds){
    return to!int(to!real(seconds) * 1000);
}

bool isdigits(string str){
    foreach(ch; str){
        if(!std.ctype.isdigit(ch)) return false;
    }
    return true;
}

bool isasciis(string str){
    foreach(ch; str){
        if(!std.ctype.isascii(ch)) return false;
    }
    return true;
}

unittest{
    assert(!maybeStringExpr("amz"));
    assert(!maybeStringExpr("AMZ"));
    assert(!maybeStringExpr("012"));
    assert(!maybeStringExpr("    "));
    assert(!maybeStringExpr("@-:;/.,"));
    assert(maybeStringExpr("'ab'"));
    assert(maybeStringExpr("ほげ"));
}

bool maybeStringExpr(string str){
    return !isasciis(str) || std.algorithm.find(str, '\'').length;
}
